#include "Sequences.h"
#include "Scene2D.h"
#include <AutoPtr.h>
#include <Log/LogPtr.h>
#include <Color.h>
#include <Format/PictureFormatManager.h>
#include "HardwarePicture2D.h"
#include <algorithm>
#include <safe_new.h>

void Sequences::Sequence::render_frame(Scene2D * scene, int fr, int alpha)const
{
	if( fr >= 0 && fr < frames.size() )
		scene->blit( frames[fr].ha, alpha );
}
int Sequences::Sequence::get_width(int fr)const
{
	return (fr < 0 || fr >= frames.size()) ? 0 : frames[fr].width;
}
int Sequences::Sequence::get_height(int fr)const
{
	return (fr < 0 || fr >= frames.size()) ? 0 : frames[fr].height;
}
float Sequences::Sequence::get_duration(int fr)const
{
	return (fr < 0 || fr >= frames.size()) ? 1 : frames[fr].duration;
}
int Sequences::Sequence::get_frame_count()const
{
	return frames.size();
}

std::list<Sequences *> * Sequences::sequences;
int Sequences::counter = 0;

class SequncesEqPred
{
	const String & name;
public:
	SequncesEqPred(const String & name)
		:	name( name )
	{}
	bool operator()(const Sequences * seq)
	{
		return seq->get_name() == name;
	}
};

int Sequences::cached_count()
{
	return sequences ? sequences->size() : 0;
}

void Sequences::unload(const Sequences * seq)
{
	if( !sequences )
	{
		LogPtr()->error( String("in Sequences::unload() sequences table is already destroyed") );
		return;
	}
	std::list<Sequences *>::iterator it = std::find( sequences->begin(), sequences->end(), seq );
	if( it == sequences->end() )
	{
		LogPtr()->warning( String("in Sequences::unload() <") + seq->get_name() + String("> not found") );
		return;
	}
	if( --(*it)->ref_count == 0 )
	{
		delete *it; *it = 0;
		it = sequences->erase( it );
	}
	if( --counter == 0 )
	{
		delete sequences; sequences = 0;
	}
}

const Sequences * Sequences::load(const String & name, Scene2D * scene)
{
	if( ++counter == 1 )
	{
		if( sequences )
		{
			LogPtr()->error( String("in Sequences::load() sequences table is already created") );
		}
		else
		{
			sequences = new std::list<Sequences *>;
		}
	}
	std::list<Sequences *>::iterator it = std::find_if( sequences->begin(), sequences->end(), SequncesEqPred( name ) );
	if( it == sequences->end() )
		it = sequences->insert( sequences->end(), new Sequences( name, scene ) );
	++(*it)->ref_count;
	return *it;
}

Sequences::Sequences(const String & name, Scene2D * scene)
:	name( name ),
	scene( scene ),
	ref_count( 0 )
{
	AutoPtr<PictureFormat> pic = PictureFormatManager::create_format( name );
	if( pic )
	{
		const Color * separator = pic->colors() + pic->get_shift( pic->width()-1, 0 );

		int start_y = 0;
		int end_y;
		do
		{
			const Color * cc = pic->colors() + pic->get_shift(0, start_y );
			while( start_y < pic->height() && *cc == *separator )
			{
				++start_y;
				cc += pic->width();
			}
			end_y = start_y;
			while( end_y < pic->height() && *cc != *separator )
			{
				++end_y;
				cc += pic->width();
			}
			if( end_y != start_y )
			{
				data.push_back( Sequence() );
				int start_x = 0;
				int end_x;
				do
				{
					const Color * cc = pic->colors() + pic->get_shift( start_x, start_y );
					while( start_x < pic->width() && *cc == *separator )
					{
						++start_x;
						++cc;
					}
					end_x = start_x;
					while( end_x < pic->width() && *cc != *separator )
					{
						++end_x;
						++cc;
					}
					if( end_x != start_x )
					{
						// Found a frame
						const Color * address = pic->colors() + pic->get_shift( start_x, start_y );
						int sx = end_x - start_x;
						int sy = end_y - start_y;
						HardwarePicture2D * ha = scene->load_picture(
								address, sx, sy, pic->width() );
						data.back().frames.push_back( Frame( 0.2f, sx, sy, ha ) );
					}
					start_x = end_x;
				}while( start_x < pic->width() );
			}
			start_y = end_y;
		}while( start_y < pic->height() );
	}
	if( data.empty() )
	{
		Color dummy[4] = { Color(255,0,0), Color(0,255,0), Color(0,255,0), Color(255,0,0) };
		data.push_back( Sequence() );
		HardwarePicture2D * ha = scene->load_picture(
							dummy, 2, 2, 2 );
		data[0].frames.push_back( Frame( 0.5f, 2, 2, ha ) );
	}
}

Sequences::~Sequences()
{
	for( int s = 0; s < data.size(); ++s )
		for( int f = 0; f < data[s].frames.size(); ++f )
		{
			delete data[s].frames[f].ha; data[s].frames[f].ha = 0;
		}
}

int Sequences::get_frame_count(int seq)const
{
	return (seq < 0 || seq >= data.size()) ? 0 : data[seq].get_frame_count();
}
int Sequences::get_height(int seq, int fr)const
{
	return (seq < 0 || seq >= data.size()) ? 0 : data[seq].get_height( fr );
}
int Sequences::get_sequence_count()const
{
	return data.size();
}
int Sequences::get_width(int seq, int fr )const
{
	return (seq < 0 || seq >= data.size()) ? 0 : data[seq].get_width( fr );
}
void Sequences::render_frame(int seq, int fr, int alpha)const
{
	if( seq >= 0 && seq < data.size() )
		data[seq].render_frame( scene, fr, alpha );
}

float Sequences::get_duration(int seq, int fr)const
{
	return (seq < 0 || seq >= data.size()) ? 1.0 : data[seq].get_duration( fr );
}
